

from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.monitor.monitor import get_current_monitor
from herre.console.context import get_current_console
from arkitekt.packers.utils import shrink_outputs
from arkitekt.schema.node import Node
from re import template
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from herre.auth import get_current_herre
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from typing import Dict
from arkitekt.messages.postman.assign.assign_return import AssignReturnMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.base import MessageModel
from arkitekt.messages.postman.provide.provide_transition import ProvideState
from arkitekt.transport.base import Transport
from arkitekt.packers.transpilers.base import Transpiler
from herre.loop import loopify
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
import asyncio
from asyncio.tasks import Task
from arkitekt.messages.postman.provide import ProvideTransitionMessage
import logging
from arkitekt.schema.template import Template



logger = logging.getLogger(__name__)


class Actor:
    transpilers = None
    expand_inputs = True
    shrink_outputs = True

    def __init__(self, provision: BouncedProvideMessage, transport: Transport, loop = None, monitor=None) -> None:
        self.provision = provision
        self.transport = transport
        self.loop = loop or get_current_herre().loop
        self.console = get_current_console()
        self.monitor = monitor or get_current_monitor()
        self.panel = self.monitor.create_actor_panel(self) if self.monitor else None
        self._template = None

        self.runningAssignments: Dict[str, asyncio.Task] = {} # Running assignments indexed by assignment reference


        self.in_queue = asyncio.Queue()
        super().__init__()

    @property
    def template(self) -> Template:
        assert self._template is not None, "We can only access the template after this Actor has been provisioned"
        return self._template


    async def log(self, message, level = LogLevel.INFO):
        if self.panel: self.panel.log(message, level=level)
        logger.info("{level}: {message}")


    async def provide_log(self, message: str, level = LogLevel.INFO):
        await self.transport.forward(ProvideLogMessage(
                data= {
                    "level": level,
                    "message": message
                },
                meta=self.provision.meta.dict(exclude={"type"})
            ))

        if self.panel: self.panel.log(message, level=level)
        logger.info(f"Provide Log: {level}: {message}")


    def run(self, *args, **kwargs):
        return loopify(self.arun(*args, **kwargs))

    async def acall(self, message: MessageModel):
        await self.in_queue.put(message)


    async def on_provide(self):
        return None


    async def on_unprovide(self):
        return None


    async def on_assign(self, message: BouncedForwardedAssignMessage):
        raise NotImplementedError("Needs to be owerwritten in Actor Subclass")


    async def arun(self):
        try:
            self._template = await Template.asyncs.get(id=self.provision.data.template)
            
            if self.panel: self.panel.start()

            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.PROVIDING,
                    "message": "We just got started Bay"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))



            await self.on_provide()

            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.ACTIVE,
                    "message": "We just got started Bay"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))

            while True:
                await self.log("Waiting for assignmements")
                message = await self.in_queue.get()
                logger.info(f"Received Message {message}")
                
                if isinstance(message, BouncedForwardedAssignMessage):
                    await self.log("Assigningment received")
                    task = self.loop.create_task(self.on_assign(message))
                    self.runningAssignments[message.meta.reference] = task

                if isinstance(message, BouncedForwardedUnassignMessage):
                    if message.data.assignation in self.runningAssignments:
                        task = self.runningAssignments[message.data.assignation]
                        if not task.done():
                            logger.info("Task is being cancelled")
                            task.cancel()
                        else:
                            logger.error("Task was already done")
                    else:
                        logger.error("Task was never assigned to this actor")
                        if self.string: raise Exception("Received cancellation for Task that was never assinged to this actor!")



        except Exception as e:
            await self.log(f"Provision Exception {str(e)}")
            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.ERROR,
                    "message": f"{e}"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))


        except asyncio.CancelledError as e:
            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.CANCELING,
                    "message": f"{e}"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))

            await self.on_unprovide()


            logger.info("Doing Whatever needs to be done to cancel!")
            await self.transport.forward(ProvideTransitionMessage(
                data= {
                    "state": ProvideState.CANCELLED,
                    "message": f"{e}"
                },
                meta={
                    "reference": self.provision.meta.reference,
                    "extensions": self.provision.meta.extensions
                }
            ))
            raise e

        if self.panel: self.panel.end()









from herre.access.object import GraphQLObject
from typing import List, Optional, Union


class Widget(GraphQLObject):
    dependencies: Optional[List[str]]

    def _repr_html_list(self):
        return f"""
        <div class="container" style="border:1px solid #00000f;padding: 4px;">
            <div class="item item-1 font-xl">{self.__class__.__name__}</div>
        </div>
        """

class ValueWidget(Widget):
    pass


class QueryWidget(Widget):
    query: Optional[str]
    list: Optional[bool]


class SearchWidget(Widget):
    query: Optional[str]

class IntWidget(Widget):
    pass

class SliderWidget(Widget):
    """ A Slider widget enables
    a slider from min to max with
    equally spaced ticks.
    Please assert default is within
    the minimum and maximum

    Args:
        max(int): The Maximum Value for this Slider
        min(int): The Minimum Value for this Slider
    """
    min: Optional[int]
    max: Optional[int]


AllWidgets = Union[SearchWidget, IntWidget, SliderWidget, QueryWidget, ValueWidget]


from abc import abstractmethod
from arkitekt.schema.widgets import AllWidgets, QueryWidget, SearchWidget, SliderWidget
from arkitekt.packers.structure import Structure
from typing import Any, Dict, List, Optional, Union
from pydantic import BaseModel, Field, validator
from herre.access.object import GraphQLObject
from arkitekt.packers.registry import get_packer_registry
from typing import ForwardRef
import asyncio




class Port(GraphQLObject):
    key: Optional[str]
    description: Optional[str]
    label: Optional[str]
    transpile: Optional[str]

    @classmethod
    def from_params(cls, widget=None, **kwargs):
        #TODO: This weird widget conversion thing needs to stop, type GraphQLOBject correctly
        port = cls(__typename=cls.__name__, widget= widget.dict() if widget else None, **kwargs) # We ensure creation of a proper object)
        return port
    

class ArgPort(Port):
    widget: Optional[AllWidgets]

    def _repr_html_list(self):
        nana =  f"""
        <div class="container" style="border:1px solid #00000f;padding: 4px;">
            <div class="item item-1 font-xl">{self.key}</div>
            <div class="item item-2">{self.description}</div>
        """
        if self.widget: nana += self.widget._repr_html_list()
        return nana + "</div>"





class IntExpandShrink:

    async def expand(self, value,**kwargs):
        return int(value)

    async def shrink(self, instance,**kwargs):
       return int(instance)


class StringExpandShrink:

    async def expand(self, value,**kwargs):
        return str(value)

    async def shrink(self, instance,**kwargs):
        return str(instance)

class StructureExpandShrink:

    async def expand(self, value, transpile=True):
        structure = get_packer_registry().get_structure(self.identifier)
        return await structure.expand(value)

    async def shrink(self, instance,**kwargs):
        if isinstance(instance, Structure) or hasattr(instance, "shrink"): return await instance.shrink()
        # Instance we are trying to shrink needs to be transpile to the required model
        from arkitekt.packers.transpilers.registry import get_transpiler_registry
        transpiler = get_transpiler_registry().get_transpiler(instance.__class__.__name__, self.identifier)
        transpiled_instance = await transpiler.transpile(instance)
        return await transpiled_instance.shrink()

class ListExpandShrink:

    async def expand(self, value,**kwargs):
        return await asyncio.gather(*[self.child.expand(item,**kwargs) for item in value])

    async def shrink(self, instance,**kwargs):
        assert isinstance(instance, list), f"ListPorts only accept lists! Got {instance}"
        return await asyncio.gather(*[self.child.shrink(item,**kwargs) for item in instance])

class DictExpandShrink:

    async def expand(self, value,**kwargs):
        return {key: await self.child.expand(item,**kwargs) for key, item in value.items()}

    async def shrink(self, instance,**kwargs):
        return {key: await self.child.shrink(item,**kwargs) for key, item in instance.items()}




# Args



class IntArgPort(ArgPort, IntExpandShrink):
    pass

class StringArgPort(ArgPort, StringExpandShrink):
    pass

class StructureArgPort(ArgPort,StructureExpandShrink):
    identifier: str

ListArgPort = ForwardRef('ListArgPort')
class ListArgPort(ListExpandShrink, ArgPort):
    child: Union[IntArgPort, StructureArgPort, StringArgPort, ListArgPort]

ListArgPort.update_forward_refs()

DictArgPort = ForwardRef('DictArgPort')
class DictArgPort(DictExpandShrink, ArgPort):
    child: Union[IntArgPort, StructureArgPort, StringArgPort, ListArgPort]

DictArgPort.update_forward_refs()


#Kwargs

class KwargPort(Port):
    widget: Optional[AllWidgets]


class IntKwargPort(KwargPort, IntExpandShrink):
    default: Optional[int]


class StringKwargPort(KwargPort, StringExpandShrink):
    default: Optional[str]

class StructureKwargPort(KwargPort, StructureExpandShrink):
    default: Optional[str]
    identifier: str

ListKwargPort = ForwardRef('ListKwargPort')
class ListKwargPort(KwargPort, ListExpandShrink):
    default: Optional[List]
    child: Union[IntKwargPort, StructureKwargPort, StringKwargPort, ListKwargPort]

ListKwargPort.update_forward_refs()

DictKwargPort = ForwardRef('DictKwargPort')
class DictKwargPort(KwargPort, DictExpandShrink):
    default: Optional[Dict]
    child: Union[IntKwargPort, StructureKwargPort, StringKwargPort, ListKwargPort, DictKwargPort]

DictKwargPort.update_forward_refs()



# Returns

class ReturnPort(Port):
    pass


class StructureReturnPort(ReturnPort, StructureExpandShrink):
    identifier: str

class IntReturnPort(ReturnPort, IntExpandShrink):
    pass

class StringReturnPort(ReturnPort, StringExpandShrink):
    pass

ListReturnPort = ForwardRef('ListReturnPort')
class ListReturnPort(ReturnPort, ListExpandShrink):
    child: Union[IntReturnPort, StructureReturnPort, StringReturnPort, ListReturnPort]

ListReturnPort.update_forward_refs()


DictReturnPort = ForwardRef('DictReturnPort')
class DictReturnPort(ReturnPort, DictExpandShrink):
    child: Union[IntReturnPort, StructureReturnPort, StringReturnPort, ListReturnPort, DictReturnPort]

DictReturnPort.update_forward_refs()




from arkitekt.monitor.monitor import Monitor
from arkitekt.actors.base import Actor
from arkitekt.messages.postman.unprovide.bounced_unprovide import BouncedUnprovideMessage
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.reserve.bounced_reserve import BouncedReserveMessage
from arkitekt.transport.base import Transport
from arkitekt.messages.base import MessageDataModel
import asyncio
from herre.loop import loopify
from typing import Callable, Dict, List, Tuple, Type
from arkitekt.transport.websocket import WebsocketTransport
from arkitekt.actors.actify import actify, define
from arkitekt.schema.node import Node
from arkitekt.schema.template import Template
from arkitekt.ward import ArkitektWard
from arkitekt.messages import ProvideLogMessage, ProvideCriticalMessage
from arkitekt.schema.negotiation import PostmanProtocol, Transcript
from herre.wards.registry import get_ward_registry
from herre.wards.base import WardException
from herre.auth import get_current_herre
from arkitekt.legacy.utils import *
from arkitekt.registry import set_current_agent

try:
    from rich.traceback import install
    install()
except:
    pass
import logging


logger = logging.getLogger(__name__)



class TemplateParams(dict):
    pass


async def parse_params(params: dict) -> TemplateParams:
    return TemplateParams(**params)


class AgentException(Exception):
    pass


    
class Agent():

    def __init__(self, auto_login=True, auto_connect=True, loop=None, register=True,  with_monitor=False, **kwargs) -> None:
        self.auto_login = auto_login
        self.auto_connect = auto_connect

        self.herre = get_current_herre(**kwargs) 
        self.loop = self.herre.loop
        
        self.ward: ArkitektWard = get_ward_registry().get_ward_instance("arkitekt")
        self.monitor = Monitor("Agent") if with_monitor else None
        self.panel = self.monitor.create_agent_panel(self) if self.monitor else None

        self.message_queue = asyncio.Queue()
        self._connected = False
        self.transport: Transport = None

        if register:
            set_current_agent(self)


        super().__init__()


    @property
    def connected(self):
        """Is this client connected

        Returns:
            [type]: [description]
        """
        return self._connected

    async def disconnect(self):
        await self.ward.disconnect()


    async def broadcast(self, message: MessageDataModel):
        raise NotImplementedError("Please Overwrite")

        
        
    async def connect(self):

        if not self.herre.logged_in:
            assert self.auto_login, "Herre is not logged in and auto_login was set to false. Please login with Herre first!"
            await self.herre.login()

        if not self.ward.connected:
            assert self.auto_connect, "We have not connected to Arkitekt before and autoconnect was set to false. Please connect a ward before or set auto_connect to True"
            await self.ward._connect()


        self.transcript = self.ward.transcript

        # Setup Postman
        agent_type = "WEBSOCKET"

        if agent_type == PostmanProtocol.WEBSOCKET:
            self.transport = WebsocketTransport(broadcast=self.broadcast, route="agent")
        
        self._connected = True

        if self.monitor: self.monitor.__enter__()




    async def aprovide(self):
        """Provide

        Provide takes registered function updates their definition, syncs up with arkitekt and makes them discoverable on arkitekt.
        It then connects as a provider to arkitekt and waits for (already registered) provisions and runs in an infinite providing
        loop.

        Raises:
            ProviderException: [description]
            ProviderException: [description]
        """
        raise NotImplementedError("Please Overwrite")


    def provide(self, as_task=False):
        return loopify(self.aprovide(), as_task=as_task)



    async def __aenter__(self):
        await self.connect()
        return self


            
    async def __aexit__(self, *args, **kwargs):
        await self.disconnect()



import asyncio
from arkitekt.messages.base import MessageModel
from arkitekt.transport.websocket import WebsocketTransport
from arkitekt.transport.base import Transport
from arkitekt.actors.actify import define
from arkitekt.schema.node import Node
from arkitekt.ward import ArkitektWard
from arkitekt.schema.negotiation import PostmanProtocol, Transcript
from herre.wards.registry import get_ward_registry
from herre.auth import get_current_herre
from arkitekt.legacy.utils import *
from arkitekt.registry import set_current_rpc

try:
    from rich.traceback import install
    install()
except:
    pass

import logging

logger = logging.getLogger(__name__)


class RPC():

    def __init__(self, auto_login=True, auto_connect=True, loop=None, register=True, **kwargs) -> None:

        self.auto_login = auto_login
        self.auto_connect = auto_connect
        self.transport: Transport = None
        self.herre = get_current_herre(**kwargs) 
        self.loop = self.herre.loop
        self.ward: ArkitektWard = get_ward_registry().get_ward_instance("arkitekt")
        self.scopes = self.herre.grant.scopes

        self._connected = False
        self.queues = {}

        if register:
            set_current_rpc(self)


        super().__init__(**kwargs)


    @property
    def connected(self):
        """Is this client connected

        Returns:
            [type]: [description]
        """
        return self._connected

    async def disconnect(self):
        await self.ward.disconnect()


    async def broadcast(self, message: MessageModel):
        if message.meta.reference in self.queues:
            logger.info(f"RPC Message Received")
            await self.queues[message.meta.reference].put(message)
        else:
            logger.error(f"Message without a reply channel {message}")


    async def stream_replies_to_queue(self, message: MessageModel):
        """ Creates a queue for this referenced message
        and forwards every message on a reply channel to the return queue

        Args:
            message (MessageModel): [description]

        Returns:
            asyncio.Queue: The Reply Queue
        """
        reply_channel = message.meta.reference
        reply_queue = asyncio.Queue()
        self.queues[reply_channel] = reply_queue
        await self.transport.forward(message)
        return reply_queue


    async def forward(self, message: MessageModel):
        await self.transport.forward(message)
        
        
    async def connect(self):

        if not self.herre.logged_in:
            assert self.auto_login, "Herre is not logged in and auto_login was set to false. Please login with Herre first!"
            await self.herre.login()

        if not self.ward.connected:
            assert self.auto_connect, "We have not connected to Arkitekt before and autoconnect was set to false. Please connect a ward before or set auto_connect to True"
            await self.ward._connect()


        self.transcript = self.ward.transcript

        # Setup Postman
        postman_type = self.transcript.postman.type

        if postman_type == PostmanProtocol.WEBSOCKET:
            self.transport = WebsocketTransport(**self.transcript.postman.kwargs, broadcast=self.broadcast, route="all")

        await self.transport.connect()
        self._connected = True


    async def __aenter__(self):
        await self.connect()
        return self


            
    async def __aexit__(self, *args, **kwargs):
        await self.disconnect()






